# ubirch protocol
#
# Copyright (c) 2018 ubirch GmbH.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import logging
from abc import abstractmethod
from uuid import UUID

import msgpack

logger = logging.getLogger(__name__)

# ubirch-protocol constants
UBIRCH_PROTOCOL_VERSION = 1

PLAIN = ((UBIRCH_PROTOCOL_VERSION << 4) | 0x01)
SIGNED = ((UBIRCH_PROTOCOL_VERSION << 4) | 0x02)
CHAINED = ((UBIRCH_PROTOCOL_VERSION << 4) | 0x03)

UBIRCH_PROTOCOL_TYPE_BIN = 0x00
UBIRCH_PROTOCOL_TYPE_REG = 0x01
UBIRCH_PROTOCOL_TYPE_HSK = 0x02


class Protocol(object):
    _signatures: dict = {}

    def __init__(self, signatures: dict = None) -> None:
        """
        Initialize the protocol.
        :param signatures: previously known signatures
        """
        if signatures is None:
            signatures = {}
        self._signatures = signatures

    def set_saved_signatures(self, signatures: dict) -> None:
        """
        Set known signatures from devices we have talked to.
        :param signatures: the saved signatures as a dictionary (uuid -> bytes)
        """
        self._signatures = signatures

    def get_saved_signatures(self) -> dict:
        """
        Get the saved signatures to store them persistently.
        :return: a dictionary of signatures (uuid -> bytes)
        """
        return self._signatures

    @abstractmethod
    def _sign(self, uuid: UUID, message: bytes) -> bytes:
        """
        Sign the request when finished.
        :param uuid: the uuid of the sender to identify the correct key pair
        :param message: the bytes to sign
        :return: the signature
        """
        raise NotImplementedError("signing not implemented")

    @abstractmethod
    def _verify(self, uuid: UUID, message: bytes, signature: bytes) -> bytes:
        """
        Verify the message. Throws exception if not verifiable.
        :param uuid: the uuid of the sender to identify the correct key pair
        :param message: the message bytes to verify
        :param signature: the signature to use for verification
        :return:
        """
        raise NotImplementedError("verification not implemented")

    def __serialize(self, msg: any) -> bytearray:
        return bytearray(msgpack.packb(msg))

    def __sign(self, uuid: UUID, msg: any) -> (bytes, bytes):
        # sign the message and store the signature
        serialized = self.__serialize(msg)[0:-1]
        sha515digest = hashlib.sha512(serialized).digest()
        signature = self._sign(uuid, sha515digest)
        # replace last element in array with the signature
        msg[-1] = signature
        return (signature, self.__serialize(msg))

    def message_signed(self, uuid: UUID, type: int, payload: any) -> bytes:
        """
        Create a new signed ubirch-protocol message.
        :param uuid: the uuid of the device that sends the message, part of the envelope
        :param type: a hint of the type of message sent (0-255)
        :param payload: the actual message payload
        :return: the encoded and signed message
        """
        # we need to ensure we get a 16bit integer serialized (0xFF | version)
        # the 0xFF is replaced by 0x00 in the serialized code
        msg = [
            SIGNED,
            uuid.bytes,
            type,
            payload,
            0
        ]

        (signature, serialized) = self.__sign(uuid, msg)
        self._signatures[uuid] = signature

        # serialize result and return the message
        return serialized

    def message_chained(self, uuid: UUID, type: int, payload: any) -> bytes:
        """
        Create a new chained ubirch-protocol message.
        Stores the context, the last signature, to be included in the next message.
        :param uuid: the uuid of the device that sends the message, part of the envelope
        :param type: a hint of the type of message sent (0-255)
        :param payload: the actual message payload
        :return: the encoded and signed message
        """

        # retrieve last known signature or null bytes
        last_signature: bytes = self._signatures.get(uuid, b'\0' * 64)

        # we need to ensure we get a 16bit integer serialized (0xFF | version)
        # the 0xFF is replaced by 0x00 in the serialized code
        msg = [
            CHAINED,
            uuid.bytes,
            last_signature,
            type,
            payload,
            0
        ]

        (signature, serialized) = self.__sign(uuid, msg)
        self._signatures[uuid] = signature

        # serialize result and return the message
        return serialized

    def message_verify(self, message: bytes) -> dict:
        """
        Verify the integrity of the message and decode the contents.
        Throws an exception if the message is not verifiable.
        :param message: the msgpack encoded message
        :return: the decoded message
        """
        if len(message) < 70:
            raise Exception("message format wrong (size < 70 bytes): {}".format(len(message)))
        unpacked = msgpack.unpackb(message)
        uuid = UUID(bytes=unpacked[1])
        if unpacked[0] == SIGNED:
            signature = unpacked[4]
        else:
            signature = unpacked[5]
        self._verify(uuid, message[0:-67], signature)
        return unpacked



# ubirch key store
#
# @author Matthias L. Jugel
#
# Copyright (c) 2018 ubirch GmbH.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timedelta
from logging import getLogger
from os import urandom
from uuid import UUID

import ed25519
from ed25519 import SigningKey, VerifyingKey
from jks import jks, AlgorithmIdentifier, rfc5208, TrustedCertEntry
from pyasn1.codec.ber import encoder

logger = getLogger(__name__)

ECC_ENCRYPTION_OID = (1, 2, 1, 3, 101, 112)


class ED25519Certificate(TrustedCertEntry):

    def __init__(self, alias: str, verifying_key: VerifyingKey, **kwargs):
        super().__init__(**kwargs)
        self.alias = alias
        self.cert = verifying_key.to_bytes()
        self.timestamp = int(datetime.utcnow().timestamp())


class KeyStore(object):
    """
    The ubirch key store handles the keys relevant for the ubirch protocol.
    """

    def __init__(self, keystore_file: str, password: str) -> None:
        """Initialize the ubirch-protocol for the device with the given UUID."""
        super().__init__()
        self._ks_file = keystore_file
        self._ks_password = password
        self._load_keys()

    def _load_keys(self) -> None:
        """Load or create new crypto-keys. The keys are stored in a local key store."""
        try:
            self._ks = jks.KeyStore.load(self._ks_file, self._ks_password)
        except FileNotFoundError:
            logger.warning("creating new key store: {}".format(self._ks_file))
            self._ks = jks.KeyStore.new("jks", [])

    def insert_ed25519_signing_key(self, uuid, sk: SigningKey):
        """Insert an existing ED25519 signing key."""
        # encode the ED25519 private key as PKCS#8
        private_key_info = rfc5208.PrivateKeyInfo()
        private_key_info.setComponentByName('version', 'v1')
        a = AlgorithmIdentifier()
        a.setComponentByName('algorithm', ECC_ENCRYPTION_OID)
        private_key_info.setComponentByName('privateKeyAlgorithm', a)
        private_key_info.setComponentByName('privateKey', sk.to_bytes())
        pkey_pkcs8 = encoder.encode(private_key_info)
        pke = jks.PrivateKeyEntry.new(alias=str(uuid.hex), certs=[], key=pkey_pkcs8)
        self._ks.entries['pke_' + uuid.hex] = pke

    def insert_ed25519_verifying_key(self, uuid, vk: VerifyingKey):
        # store verifying key in certificate store
        self._ks.entries[uuid.hex] = ED25519Certificate(uuid.hex, vk)

    def insert_ed25519_keypair(self, uuid: UUID, vk: VerifyingKey, sk: SigningKey) -> (VerifyingKey, SigningKey):
        """Insert an existing ED25519 key pair into the key store."""
        if uuid.hex in self._ks.entries or uuid.hex in self._ks.certs:
            raise Exception("uuid '{}' already exists in keystore".format(uuid.hex))

        self.insert_ed25519_verifying_key(uuid, vk)
        self.insert_ed25519_signing_key(uuid, sk)
        self._ks.save(self._ks_file, self._ks_password)
        logger.info("created new key pair for {}: {}".format(uuid.hex, bytes.decode(vk.to_ascii(encoding='hex'))))
        return (vk, sk)

    def create_ed25519_keypair(self, uuid: UUID) -> (VerifyingKey, SigningKey):
        """Create a new ED25519 key pair and store in key store."""
        sk, vk = ed25519.create_keypair(entropy=urandom)
        return self.insert_ed25519_keypair(uuid, vk, sk)

    def exists_signing_key(self, uuid: UUID):
        """Check whether this UUID has a signing key in the key store."""
        return 'pke_' + uuid.hex in self._ks.private_keys

    def exists_verifying_key(self, uuid: UUID):
        """Check whether this UUID has a verifying key in the key store."""
        return uuid.hex in self._ks.certs

    def find_signing_key(self, uuid: UUID) -> SigningKey:
        """Find the signing key for this UUID."""
        sk = self._ks.private_keys['pke_' + uuid.hex]
        return SigningKey(sk.pkey)

    def find_verifying_key(self, uuid: UUID) -> VerifyingKey:
        """Find the verifying key for this UUID."""
        cert = self._ks.certs[uuid.hex]
        return VerifyingKey(cert.cert)

    def get_certificate(self, uuid: UUID) -> dict or None:
        if not uuid.hex in self._ks.certs:
            return None

        cert = self._ks.certs[uuid.hex]
        vk = VerifyingKey(cert.cert)
        created = datetime.fromtimestamp(cert.timestamp)
        not_before = datetime.fromtimestamp(cert.timestamp)
        # TODO fix handling of key validity
        not_after = created + timedelta(days=365)
        return {
            "algorithm": 'ECC_ED25519',
            "created": int(created.timestamp()),
            "hwDeviceId": uuid.bytes,
            "pubKey": vk.to_bytes(),
            "pubKeyId": vk.to_bytes(),
            "validNotAfter": int(not_after.timestamp()),
            "validNotBefore": int(not_before.timestamp())
        }



# ubirch API
#
# @author Matthias L. Jugel
#
# Copyright (c) 2018 ubirch GmbH.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import binascii
import json
import logging
from logging import getLogger
from uuid import UUID

import requests
from requests import Response

logger = getLogger(__name__)


class API(object):
    """ubirch API accessor methods."""

    def __init__(self, auth=None, env=None, debug=False) -> None:
        super().__init__()

        # enable intensive logging
        if debug and logger.level == logging.DEBUG:
            import http.client as http_client
            http_client.HTTPConnection.debuglevel = 1
            requests_log = logging.getLogger("requests.packages.urllib3")
            requests_log.setLevel(logging.DEBUG)
            requests_log.propagate = True

        if auth is not None:
            self._auth = {'Authorization': auth}
        else:
            self._auth = {}

        if env is None:
            self.KEY_SERVICE = "http://localhost:8095/api/keyService/v1"
            self.AVATAR_SERVICE = "http://localhost:8080/api/avatarService/v1"
            self.CHAIN_SERVICE = "http://localhost:8097/api/v1/chainService"
            self.NOTARY_SERVICE = "https://localhost:8098/api/v1/notaryService"
        else:
            self.KEY_SERVICE = "https://key.{}.ubirch.com/api/keyService/v1".format(env)
            self.AVATAR_SERVICE = "https://api.ubirch.{}.ubirch.com/api/avatarService/v1".format(env)
            self.CHAIN_SERVICE = "https://api.ubirch.{}.ubirch.com/api/v1/chainService".format(env)
            self.NOTARY_SERVICE = "http://n.dev.ubirch.com:8080/v1/notaryService".format(env)

    def is_identity_registered(self, uuid: UUID) -> bool:
        """
        Check if this identity is registered with the backend.
        :param uuid: the UUID of the identity to check
        :return: true if the identity exists
        """
        logger.info("is identity registered?: {}".format(uuid))
        r = requests.get(self.KEY_SERVICE + "/pubkey/current/hardwareId/" + str(uuid),
                         headers=self._auth)
        logger.debug("{}: {}".format(r.status_code, r.content))
        return r.status_code == 200 and r.json()

    def register_identity(self, key_registration: bytes) -> Response:
        """
        Register an identity with the backend.
        :param key_registration: the key registration data
        :return: the response from the server
        """
        if key_registration.startswith(b'{'):
            logger.debug(key_registration)
            return self._register_identity_json(json.loads(bytes.decode(key_registration)))
        else:
            return self._register_identity_mpack(key_registration)

    def _register_identity_json(self, key_registration: dict) -> Response:
        logger.info("register device identity [json]: {}".format(key_registration))
        r = requests.post(self.KEY_SERVICE + '/pubkey', json=key_registration,
                          headers=self._auth)
        logger.debug("{}: {}".format(r.status_code, r.content))
        return r

    def _register_identity_mpack(self, key_registration: bytes) -> Response:
        logger.info("register device identity [msgpack]: {}".format(binascii.hexlify(key_registration)))
        r = requests.post(self.KEY_SERVICE + '/pubkey/mpack', data=key_registration,
                          headers={'Content-Type': 'application/octet-stream', **self._auth})
        logger.debug("{}: {}".format(r.status_code, r.content))
        return r

    def device_exists(self, uuid: UUID) -> bool:
        """
        Check if a device exists.
        :param uuid: the UUID of the device
        :return: true of it exists
        """
        logger.info("device exists?: {}".format(uuid))
        r = requests.get(self.AVATAR_SERVICE + '/device/' + str(uuid),
                         headers=self._auth)
        logger.debug("{}: {}".format(r.status_code, r.content))
        return r.status_code == 200

    def device_delete(self, uuid: UUID) -> bool:
        """
        Delete a device
        :param uuid: the UUID of the device
        :return: true of the deletion succeeded
        """
        logger.info("delete device: {}".format(uuid))
        r = requests.delete(self.AVATAR_SERVICE + '/device/' + str(uuid), headers=self._auth)
        logger.debug("{}: {}".format(r.status_code, r.content))
        return r.status_code == 200

    def device_create(self, device_info: dict) -> Response:
        """
        Create a new device in the server using the device info provided.
        :param device_info: a device descriptor
        :return: the response from the server
        """
        logger.info("create device: {}".format(device_info))
        r = requests.post(self.AVATAR_SERVICE + '/device',
                          json=device_info,
                          headers=self._auth)
        logger.debug("{}: {}".format(r.status_code, r.content))
        return r

    def send(self, data: bytes) -> Response:
        """
        Send data to the backend. Requires encoding before sending.
        :param data: the msgpack or JSON encoded data to send
        :return: the response from the server
        """
        if data.startswith(b'{'):
            return self._send_json(json.loads(bytes.decode(data)))
        else:
            return self._send_mpack(data)

    def anchor(self, data: bytes) -> Response:
        """
        Anchor some data in the blockchain service.
        :param data: the data to anchor
        :return: the response from the server
        """
        if data.startswith(b'{'):
            raise Exception("unsupported data type: json")

        r = requests.post(self.NOTARY_SERVICE + '/notarize',
                          json={"data": bytes.decode(binascii.hexlify(data[-64:])), "dataIsHash": True},
                          headers=self._auth)
        logger.debug("{}: {}".format(r.status_code, r.content))
        return r

    def _send_json(self, data: dict) -> Response:
        payload = str.encode(json.dumps(data, sort_keys=True, separators=(',', ':')))
        logger.debug(json)
        r = requests.post(self.AVATAR_SERVICE + '/device/update',
                          headers={'Content-Type': 'application/json'},
                          data=payload)
        logger.debug("{}: {}".format(r.status_code, r.content))
        return r

    def _send_mpack(self, data: bytes) -> Response:
        logger.debug(data)
        r = requests.post(self.AVATAR_SERVICE + '/device/update/mpack', data=data)
        logger.debug("{}: {}".format(r.status_code, r.content))
        return r



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ubirch-protocol",
    version="1.0.0",
    author="Matthias L. Jugel",
    author_email="matthias.jugel@ubirch.com",
    description="A ubirch-protocol implementation for python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ubirch/ubirch-protocol-python",
    packages=setuptools.find_packages(exclude=['bin', 'docs', 'examples', 'tests*']),
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ),
)